﻿using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using UnityEngine.UI;
using System.Linq;
using Battlehub.Utils;
using SimpleJSON;
using MVR.FileManagement;
using System.Text;
using UnityEngine.Windows.Speech;
using Battlehub.RTSaveLoad;
//using static Leap.Unity.MultiTypedList;
//using static PrefabEvolution.BaseExposedData;

namespace MVRPlugin {
    public class AutoHotKeys: MVRScript {
        public static AutoHotKeys SessionPlugin { get; private set; }

        #region Consts
        private const string _storeTriggerId = "Trigger";
        private const string _uiButton = "UIButton";
        private const string _uiSlider = "UISlider";
        private const string _uiToggle = "UIToggle";
        private const KeyCode _mainKeyCode = KeyCode.LeftControl;
        private const KeyCode _hotStoreKeyCode = KeyCode.LeftAlt;
        private const KeyCode _hotSpeechKeyCode = KeyCode.Return;
        #endregion

        #region UI
        private GameObject _vamOverlaysGO;
        private Canvas _overlaysCanvas;
        private Text _currentCommandText;
        private Text _commandsListText;
        private Text _helpTxt;
        private bool _sessionMode;
        /// <summary>
        /// Selected current command
        /// </summary>
        private Atom _command;
        /// <summary>
        /// Keyboard helper
        /// </summary>
        private Keys _keys;
        /// <summary>
        /// Type string buffer
        /// </summary>
        private string _buffer = "";
        /// <summary>
        /// Offset in command list
        /// </summary>
        private int _index = 0;
        private bool _isDirty = false;
        #endregion

        readonly Dictionary<KeyCode, string> _hotKeys = new Dictionary<KeyCode, string>();
        readonly HashSet<string> _blackList = new HashSet<string>( StringComparer.OrdinalIgnoreCase );
        readonly Dictionary<string, string> _speechHotKeys = new Dictionary<string, string>();
        Dictionary<string, string> _autoSpeechHotKeys = new Dictionary<string, string>();
        KeywordRecognizer _keywordRecognizer;

        JSONStorableString _sceneHotKeys;
        JSONStorableString _sceneSpeechHotKeys;
        JSONStorableString _sceneBlacklist;

        JSONStorableStringChooser _sideChooser;
        JSONStorableFloat _fontSize;
        JSONStorableColor _cmdColor;
        JSONStorableColor _listColor;
        JSONStorableStringChooser _comparer;
        //JSONStorableString _welcome;
        JSONStorableBool _autoSpeechMode;
        JSONStorableBool _showLogs;
        JSONStorableBool _showHelp;

        CompareDelegate _currentCompare = compare1;
        bool _hideShow = false;
        int _mode = 0;

        public override void Init() {
            try {
                List<string> items = new List<string>() {
                    "Left",
                    "Middle",
                    "Right"
                };
                List<string> compares = new List<string>() {
                    "Standart",
                    "Fuzzy",
                };
                List<string> keys = new List<string>();
                foreach( var k in Enum.GetValues( typeof( KeyCode ) ) )
                    keys.Add( k.ToString() );

                keys.Remove( "None" );
                keys.Remove( "F1" );
                keys.Remove( "F11" );
                keys.Remove( "F12" );
                keys.Remove( "A" );
                keys.Remove( "S" );
                keys.Remove( "W" );
                keys.Remove( "D" );
                keys.Remove( "UpArrow" );
                keys.Remove( "DownArrow" );



                //_welcome = new JSONStorableString( "welcome", "...", ( string a ) => { } );
                //RegisterString( _welcome );
                //CreateTextField( _welcome );

                _autoSpeechMode = new JSONStorableBool( "Auto speech mode", false, ( bool a ) => { } );
                RegisterBool( _autoSpeechMode );
                if( PhraseRecognitionSystem.isSupported )
                    CreateToggle( _autoSpeechMode );
                else
                    _autoSpeechMode.val = false;

                _showLogs = new JSONStorableBool( "Show speech text", false, ( bool a ) => { } );
                RegisterBool( _showLogs );
                CreateToggle( _showLogs );

                _showHelp = new JSONStorableBool( "Show tips", true, setBool );
                RegisterBool( _showHelp );
                CreateToggle( _showHelp );

                _sideChooser = new JSONStorableStringChooser( "side", items, "Middle", "Display side", setSide );
                RegisterStringChooser( _sideChooser );
                CreatePopup( _sideChooser );

                //_mainKey = new JSONStorableStringChooser( "mainKey", keys, "LeftControl", "Main key", UpdateKey );
                //RegisterStringChooser( _mainKey );
                //CreateFilterablePopup( _mainKey );

                //_hotStoreKey = new JSONStorableStringChooser( "hotStoreKey", keys, "LeftAlt", "Store key", UpdateKey );
                //RegisterStringChooser( _hotStoreKey );
                //CreateFilterablePopup( _hotStoreKey );

                _comparer = new JSONStorableStringChooser( "comparer", compares, "Standart", "Filter function", updateComparer );
                RegisterStringChooser( _comparer );
                CreatePopup( _comparer );

                _fontSize = new JSONStorableFloat( "fontSize", 12f, setFontSize, 5f, 50f, true, true );
                RegisterFloat( _fontSize );
                CreateSlider( _fontSize, true );

                _cmdColor = new JSONStorableColor( "cmdColor", HSVColorPicker.RGBToHSV( 0.8f, 0.8f, 0.8f ), setColor );
                RegisterColor( _cmdColor );
                CreateColorPicker( _cmdColor, true );

                _listColor = new JSONStorableColor( "listColor", HSVColorPicker.RGBToHSV( 1f, 1f, 1f ), setColor );
                RegisterColor( _listColor );
                CreateColorPicker( _listColor, true );

                //_selectedColor = new JSONStorableColor( "selectedColor", HSVColorPicker.RGBToHSV( 1f, 0f, 1f ), SetColor );
                //RegisterColor( _selectedColor );
                //CreateColorPicker( _selectedColor );

                if( PhraseRecognitionSystem.isSupported ) {
                    CreateTextField( new JSONStorableString( "x1", "<color=#00FF00><b>Speech recognition is supported on your machine!</b></color>" ) )
                        .height = 20;
                } else {
                    CreateTextField( new JSONStorableString( "x1", "<color=#FF0000><b>Apparently speech recognition is not supported on your machine!</b></color>" ) );
                }

                if( containingAtom.type == "SessionPluginManager" ) {
                    _sessionMode = true;
                    if( SessionPlugin != null ) {
                        SuperController.LogError( "Plugin already added." );
                        return;
                    }

                    SessionPlugin = this;
                } else {
                    if( SessionPlugin != null )
                        SessionPlugin.enabled = false;

                    _sceneHotKeys = new JSONStorableString( "sceneHotKeys", "" );
                    RegisterString( _sceneHotKeys );
                    _sceneSpeechHotKeys = new JSONStorableString( "sceneSpeechHotKeys", "" );
                    RegisterString( _sceneSpeechHotKeys );
                    _sceneBlacklist = new JSONStorableString( "sceneBlacklist", "" );
                    RegisterString( _sceneBlacklist );
                }

                SuperController.singleton.onSceneLoadedHandlers += load;

                _keys = new Keys();

                createTextOverlay();
                applyTextOptions();
                load();
                //applyKeys();
                updateComparer( _comparer.val );
                updateKeywordRecognizer();
            } catch( Exception e ) {
                SuperController.LogError( "Exception caught: " + e );
            }
        }

        private void createTextOverlay() {
            // Creation of the main Canvas
            _vamOverlaysGO = new GameObject( "AutoHotkeyCanvas" ) {
                layer = 5
            };
            _vamOverlaysGO.transform.SetParent( SuperController.singleton.centerCameraTarget.targetCamera.transform, false );
            _overlaysCanvas = _vamOverlaysGO.AddComponent<Canvas>();
            _overlaysCanvas.renderMode = RenderMode.ScreenSpaceOverlay;
            //_overlaysCanvas.sortingOrder = 2;
            //_overlaysCanvas.planeDistance = 10.0f;
            _vamOverlaysGO.AddComponent<CanvasScaler>();

            #region Command list text field
            // Creation of the structure for subtitles
            var _commandListTxtGO = new GameObject( "AutoHotKeySubtitlesText" ) {
                layer = 5
            };
            _commandListTxtGO.transform.SetParent( _vamOverlaysGO.transform, false );
            _commandListTxtGO.AddComponent<CanvasRenderer>();
            _currentCommandText = _commandListTxtGO.AddComponent<Text>();
            _currentCommandText.font = (Font)Resources.GetBuiltinResource( typeof( Font ), "Arial.ttf" );
            _currentCommandText.raycastTarget = false;
            _currentCommandText.text = "";
            _currentCommandText.supportRichText = true;
            // And finally again, RectTransform tweaks
            var _commandListRecTr = _currentCommandText.GetComponent<RectTransform>();
            _commandListRecTr.localPosition = new Vector3( 0, 0, 0 );
            _commandListRecTr.localScale = new Vector3( 1, 1, 1 );
            _commandListRecTr.localRotation = Quaternion.identity;
            _commandListRecTr.anchorMin = new Vector2( 0.0f, 0.0f );
            _commandListRecTr.anchorMax = new Vector2( 1f, 1f );
            _commandListRecTr.offsetMin = new Vector2( 0, 0 );
            _commandListRecTr.offsetMax = new Vector2( 0, 0 );
            #endregion

            #region Help text field
            var _helpTextGO = new GameObject( "AutoHotKeyHelpText" ) {
                layer = 5
            };
            _helpTextGO.transform.SetParent( _vamOverlaysGO.transform, false );
            _helpTextGO.AddComponent<CanvasRenderer>();
            _helpTxt = _helpTextGO.AddComponent<Text>();
            _helpTxt.font = (Font)Resources.GetBuiltinResource( typeof( Font ), "Arial.ttf" );
            _helpTxt.raycastTarget = false;
            _helpTxt.text = "HELP TEXT";
            _helpTxt.supportRichText = true;
            var _helpRecTr = _helpTxt.GetComponent<RectTransform>();
            _helpRecTr.localPosition = new Vector3( 0, 0, 0 );
            _helpRecTr.localScale = new Vector3( 1, 1, 1 );
            _helpRecTr.localRotation = Quaternion.identity;
            _helpRecTr.anchorMin = new Vector2( 0.0f, 0.0f );
            _helpRecTr.anchorMax = new Vector2( 1f, 1f );
            _helpRecTr.offsetMin = new Vector2( 0, 0 );
            _helpRecTr.offsetMax = new Vector2( 0, 0 );
            #endregion


            var _cmdsTxtGO = new GameObject( "AutoHotKeyCommandsText" ) {
                layer = 5
            };
            _cmdsTxtGO.transform.SetParent( _vamOverlaysGO.transform, false );
            _cmdsTxtGO.AddComponent<CanvasRenderer>();
            _commandsListText = _cmdsTxtGO.AddComponent<Text>();
            _commandsListText.font = (Font)Resources.GetBuiltinResource( typeof( Font ), "Arial.ttf" );
            _commandsListText.raycastTarget = false;
            _commandsListText.text = "";
            _commandsListText.supportRichText = true;

            var shadow = _cmdsTxtGO.AddComponent<Shadow>();
            shadow.effectColor = Color.black;
            shadow = _cmdsTxtGO.AddComponent<Shadow>();
            shadow.effectColor = Color.black;
            shadow.effectDistance = new Vector2( -1, 1 );
            shadow = _cmdsTxtGO.AddComponent<Shadow>();
            shadow.effectColor = Color.black;
            shadow.effectDistance = new Vector2( -1, -1 );

            // And finally again, RectTransform tweaks
            var _cmdsRecTr = _commandsListText.GetComponent<RectTransform>();
            _cmdsRecTr.localPosition = new Vector3( 0, 0, 0 );
            _cmdsRecTr.localScale = new Vector3( 1, 1, 1 );
            _cmdsRecTr.localRotation = Quaternion.identity;
            _cmdsRecTr.anchorMin = new Vector2( 0.0f, 0.0f );
            _cmdsRecTr.anchorMax = new Vector2( 1f, 1f );

            _cmdsRecTr.offsetMax = new Vector2( 1, 1 );
        }

        int _checkTicks = 0;
        string _hashString = "";

        protected void Update() {
            try {
                if( _autoSpeechMode.val ) {
                    if( ++_checkTicks > 10 ) {
                        _checkTicks = 0;
                        string hash = "";
                        var atoms = getButtonAtoms();
                        foreach( var a in atoms )
                            hash += a.Value;

                        if( hash != _hashString ) {
                            _hashString = hash;
                            _autoSpeechHotKeys = atoms;
                            updateKeywordRecognizer();
                        }
                    }
                }

                _keys.Update();

                if( _keys.IsPressed( _mainKeyCode ) )
                    _index = -1;

                if( Input.GetKey( _mainKeyCode ) ) {
                    if( _mode == 1 ) {
                        editSpeechHotKey();
                        return;
                    }

                    var atoms = getAtoms();

                    string _old = _buffer;

                    if( !Input.GetKey( _hotStoreKeyCode ) ) {
                        editFilter();
                    }

                    string selected;
                    string cmds = updateCommandList( atoms, _buffer, out _command, out selected );
                    if( cmds == "" ) {
                        _buffer = _old;
                        cmds = updateCommandList( atoms, _buffer, out _command, out selected );
                    }

                    _commandsListText.text = cmds;

                    if( _command != null ) {
                        // switch mode to edit speech hot key
                        if( PhraseRecognitionSystem.isSupported && Input.GetKeyDown( _hotSpeechKeyCode ) ) {
                            _mode = 1;
                            _commandsListText.text = "Type text for command: " + selected;
                          
                            _buffer = "";
                            return;
                        }
                    }

                    if( Input.GetKey( _hotStoreKeyCode ) ) {
                        editHotKey( ref selected );
                    } else {
                        if( _command != null ) {
                            if( _keys.IsPressed( KeyCode.Delete ) ) {
                                if( _blackList.Add( _command.uid ) ) {
                                    _isDirty = true;
                                    sceneSave();
                                }
                            } else if( _keys.IsPressed( KeyCode.LeftArrow ) ) {
                                invokeSlideCommand( -1 );
                            } else if( _keys.IsPressed( KeyCode.RightArrow ) ) {
                                invokeSlideCommand( +1 );
                            }
                        }

                        if( _isDirty ) {
                            _helpTxt.text = "PRESS PAGE DOWN TO SAVE";
                        } else {
                            _helpTxt.text = "";
                        }

                        if( string.IsNullOrEmpty( _buffer ) )
                            _currentCommandText.text = "...";
                        else
                            _currentCommandText.text = _buffer;

                        if( _keys.IsPressed( KeyCode.PageDown ) ) {
                            saveNow();
                        }

                        if( _keys.IsPressed( KeyCode.F11 ) ) {
                            _hideShow = !_hideShow;
                            if( _hideShow )
                                showAtoms();
                            else
                                hideAtoms();
                        }

                        if( _keys.IsPressed( KeyCode.F12 ) )
                            dumpAtoms();
                    }
                } else {
                    if( _mode == 1 ) {
                        _mode = 0;
                        _buffer = "";
                        _command = null;
                        _index = -1;
                    }

                    if( _keys.FirstKey != KeyCode.None ) {
                        string id;
                        if( _hotKeys.TryGetValue( _keys.FirstKey, out id ) ) {
                            _command = SuperController.singleton.GetAtomByUid( id );
                            invokeButtonCommand( true );
                        }
                    } else {
                        invokeButtonCommand( false );
                    }
                    _helpTxt.text = "";
                }
            } catch( Exception e ) {
                SuperController.LogError( "Exception caught: " + e );
            }
        }

        protected void OnDestroy() {
            // Removing the main object, every children will be destroyed too
            if( _hideShow )
                showAtoms();
            Destroy( _vamOverlaysGO );
            SuperController.singleton.onSceneLoadedHandlers -= load;
            if( _sessionMode )
                SessionPlugin = null;
            else if( SessionPlugin != null )
                SessionPlugin.enabled = true;
            destroyRecognizer();
        }

        private string updateCommandList( List<NamedAtom> atoms, string buffer, out Atom command, out string selected ) {
            string cmds = "";
            command = null;
            selected = "";

            if( buffer == "" ) {
                if( _index < -1 ) _index = -1;
            } else {
                if( _index < 0 ) _index = 0;
            }
            if( _index >= atoms.Count ) _index = atoms.Count - 1;

            int localIndex = 0;

            for( int j = 0; j < atoms.Count; ++j ) {
                var atom = atoms[j];

                if( _buffer == "" || _currentCompare( buffer, atom.name ) ) {
                    if( cmds != "" )
                        cmds += "\n";

                    if( _index == localIndex ) {
                        command = atom.atom;
                        selected = atom.name;
                        cmds += "<color=#ff00ffff>" + atom.name + "</color>";
                    } else {
                        cmds += atom.name;
                    }

                    if( !string.IsNullOrEmpty( atom.postfix ) )
                        cmds += $" ({atom.postfix})";

                    var i = getAssignedHotKeys( atom.atom );
                    if( i != "" )
                        cmds += " <color=#ff889988>[" + i + "]</color>";

                    localIndex++;
                }
            }

            if( localIndex <= _index ) _index--;

            return cmds;
        }

        private string getAssignedHotKeys( Atom a ) {
            if( a == null ) return "";

            string s = "";
            foreach( var i in _hotKeys )
                if( i.Value == a.uid ) {
                    if( s == "" )
                        s = i.Key.ToString();
                    else
                        s += ", " + i.Key.ToString();
                }

            foreach( var i in _speechHotKeys )
                if( i.Value == a.uid ) {
                    if( s == "" )
                        s = "'" + i.Key.ToString() + "'";
                    else
                        s += ", '" + i.Key.ToString() + "'";
                }

            return s;
        }

        #region UI update events
        private void updateKeywordRecognizer() {
            destroyRecognizer();

            if( PhraseRecognitionSystem.isSupported ) {
                if( _autoSpeechMode.val ) {
                    if( _autoSpeechHotKeys.Count == 0 ) return;
                    _keywordRecognizer = new KeywordRecognizer( _autoSpeechHotKeys.Keys.ToArray() );
                } else {
                    if( _speechHotKeys.Count == 0 ) return;
                    _keywordRecognizer = new KeywordRecognizer( _speechHotKeys.Keys.ToArray() );
                }

                _keywordRecognizer.OnPhraseRecognized += keywordRecognizer_OnPhraseRecognized;
                _keywordRecognizer.Start();
                if( _showLogs.val )
                    SuperController.LogMessage( "PhraseRecognitionSystem started" );
            }
        }

        private void destroyRecognizer() {
            if( _keywordRecognizer != null ) {
                _keywordRecognizer.OnPhraseRecognized -= keywordRecognizer_OnPhraseRecognized;
                _keywordRecognizer.Stop();
                _keywordRecognizer.Dispose();
                _keywordRecognizer = null;
            }
        }

        private void keywordRecognizer_OnPhraseRecognized( PhraseRecognizedEventArgs args ) {
            if( _showLogs.val )
                SuperController.LogMessage( args.text );

            string id;
            if( _speechHotKeys.TryGetValue( args.text, out id ) ) {
                _command = SuperController.singleton.GetAtomByUid( id );
                invokeButtonCommand( true );
            } else if( _autoSpeechHotKeys.TryGetValue( args.text, out id ) ) {
                if( id.EndsWith( "__" ) ) {
                    _command = SuperController.singleton.GetAtomByUid( id.Substring( 0, id.Length - 2 ) );
                    invokeSlideCommand( -100 );
                } else if( id.EndsWith( "**" ) ) {
                    _command = SuperController.singleton.GetAtomByUid( id.Substring( 0, id.Length - 2 ) );
                    invokeSlideCommand( 100 );
                } else if( id.EndsWith( "++" ) ) {
                    _command = SuperController.singleton.GetAtomByUid( id.Substring(0, id.Length - 2 ) );
                    invokeSlideCommand( 1 );
                } else if( id.EndsWith( "--" ) ) {
                    _command = SuperController.singleton.GetAtomByUid( id.Substring(0, id.Length - 2 ) );
                    invokeSlideCommand( -1 );
                } else {
                    _command = SuperController.singleton.GetAtomByUid( id );
                    invokeButtonCommand( true );
                };
            };
        }

        private void updateComparer( string key ) {
            switch( key ) {
                case "Fuzzy":
                    _currentCompare = compare2;
                    break;

                default:
                    _currentCompare = compare1;
                    break;
            }
        }

        private void setColor( float h, float s, float v ) {
            applyTextOptions();
        }

        private void setFontSize( float value ) {
            applyTextOptions();
        }

        private void setSide( string side ) {
            applyTextOptions();
        }

        private void setBool( bool b ) {
            applyTextOptions();
        }

        private void applyTextOptions() {
            switch( _sideChooser.val ) {
                case "Left":
                    _currentCommandText.alignment = TextAnchor.LowerLeft;
                    _commandsListText.alignment = TextAnchor.LowerLeft;
                    _helpTxt.alignment = TextAnchor.UpperLeft;
                    break;
                case "Right":
                    _currentCommandText.alignment = TextAnchor.LowerRight;
                    _commandsListText.alignment = TextAnchor.LowerRight;
                    _helpTxt.alignment = TextAnchor.UpperRight;
                    break;

                default:
                    _currentCommandText.alignment = TextAnchor.LowerCenter;
                    _commandsListText.alignment = TextAnchor.LowerCenter;
                    _helpTxt.alignment = TextAnchor.UpperCenter;
                    break;
            }
            _currentCommandText.fontSize = (int)_fontSize.val;
            _currentCommandText.color = HSVColorPicker.HSVToRGB( _listColor.val );
            _commandsListText.fontSize = (int)_fontSize.val;
            _commandsListText.color = HSVColorPicker.HSVToRGB( _cmdColor.val );
            _helpTxt.fontSize = (int)(_fontSize.val*2);
            _helpTxt.color = HSVColorPicker.HSVToRGB( _cmdColor.val );
            _helpTxt.gameObject.SetActive( _showHelp.val );

            var _cmdsRecTr = _commandsListText.GetComponent<RectTransform>();
            _cmdsRecTr.offsetMin = new Vector2( 0, _fontSize.val );
        }
        #endregion

        #region Commands
        private void dumpAtoms() {
            //foreach( Atom atom in SuperController.singleton.GetAtoms() ) {
            //    if( (atom.type == _uiButton || atom.type == _uiSlider || atom.type == _uiToggle) )
            //        if( atom.type.StartsWith( "UI" ) ) {
            //            SuperController.LogMessage( $"{atom.name} | {atom.type}" );
            //            foreach( var s in atom.GetStorableIDs() ) {
            //                var item = atom.GetStorableByID( s );
            //                SuperController.LogMessage( $"  {s}: {item.GetType().FullName}" );
            //            }
            //        }
            //}
            updateKeywordRecognizer();
        }

        private void editHotKey( ref string selected ) {
            if( _command == null ) {
                if( _keys.FirstKey != KeyCode.None ) {
                    _hotKeys.Remove( _keys.FirstKey );
                    setDirty();
                    selected = "";
                }

                if( _hotKeys.Count > 0 ) {
                    _helpTxt.text = "PRESS HOT KEY TO REMOVE IT";
                } else
                    _helpTxt.text = "";
                _currentCommandText.text = "";
            } else {
                if( _keys.FirstKey != KeyCode.None && _keys.FirstKey != _hotStoreKeyCode ) {
                    _hotKeys[_keys.FirstKey] = _command.uid;
                    setDirty();
                    selected = "";
                }

                _currentCommandText.text = _buffer;
                _helpTxt.text = "PRESS ANY KEY TO BECOME A HOT KEY";
            }
        }

        private void editFilter() {
            if( _keys.IsPressed( KeyCode.UpArrow ) )
                _index--;
            else if( _keys.IsPressed( KeyCode.DownArrow ) )
                _index++;
            else if( _keys.IsPressed( KeyCode.Home ) ) {
                if( _blackList.Count > 0 ) {
                    _blackList.Clear();
                    _isDirty = true;
                    sceneSave();
                }
            } else if( _keys.IsPressed( KeyCode.Backspace ) ) {
                if( _buffer.Length > 0 )
                    _buffer = _buffer.Substring( 0, _buffer.Length - 1 );
            } else if( _keys.Letter != null ) {
                _buffer += _keys.Letter;
            }
        }

        private void editSpeechHotKey() {
            if( _keys.IsPressed( KeyCode.Backspace ) ) {
                if( _buffer.Length > 0 )
                    _buffer = _buffer.Substring( 0, _buffer.Length - 1 );
            } else if( _keys.Letter != null ) {
                _buffer += _keys.Letter;
            }

            _helpTxt.text = "Type word to speech hot key and then press enter or release ctrl to cancel";

            if( Input.GetKeyDown( KeyCode.Return ) ) {
                _speechHotKeys[_buffer] = _command.uid;
                _mode = 0;
                setDirty();
                updateKeywordRecognizer();
            }

            _currentCommandText.text = _buffer;
        }    

        private void setDirty() {
            _buffer = "";
            _command = null;
            _index = -1;
            _isDirty = true;
            sceneSave();
        }

        private void invokeSlideCommand( int direction ) {
            if( _command != null ) {
                switch( _command.type ) {
                    case _uiButton: {
                            var trigger = _command.GetStorableByID( _storeTriggerId ) as UIButtonTrigger;
                            if( trigger != null && trigger.button != null && trigger.isActiveAndEnabled )
                                trigger.button.onClick.Invoke();
                        }
                        break;
                    case _uiSlider: {
                            var trigger = _command.GetStorableByID( _storeTriggerId ) as UISliderTrigger;
                            if( trigger != null && trigger.slider != null && trigger.isActiveAndEnabled ) {
                                trigger.slider.normalizedValue += direction * 0.1f;
                            }
                        }
                        break;
                    case _uiToggle: {
                            var trigger = _command.GetStorableByID( _storeTriggerId ) as UIToggleTrigger;
                            if( trigger != null && trigger.toggle != null && trigger.isActiveAndEnabled ) {
                                trigger.toggle.isOn = !trigger.toggle.isOn;
                            }
                        }
                        break;
                }
            }
        }

        private void invokeButtonCommand( bool allowToggle ) {
            if( _command != null ) {
                if( _command.type == _uiButton ) {
                    var trigger = _command.GetStorableByID( _storeTriggerId ) as UIButtonTrigger;
                    if( trigger != null && trigger.button != null && trigger.isActiveAndEnabled )
                        trigger.button.onClick.Invoke();
                } else if( allowToggle && _command.type == _uiToggle ) {
                    var trigger = _command.GetStorableByID( _storeTriggerId ) as UIToggleTrigger;
                    if( trigger != null && trigger.toggle != null && trigger.isActiveAndEnabled )
                        trigger.toggle.isOn = !trigger.toggle.isOn;
                }
                _command = null;
            }
            _buffer = "";
            _currentCommandText.text = "";
            _commandsListText.text = "";
        }

        private void hideAtoms() {
            List<Atom> atoms = SuperController.singleton.GetAtoms();
            foreach( Atom atom in atoms ) {
                if( (atom.type == _uiButton || atom.type == _uiSlider || atom.type == _uiToggle) ) {
                    var col = atom.GetStorableByID( "control" ) as FreeControllerV3;
                    if( col != null )
                        col.MoveYPositionRelativeNoForce( 10000 );
                }
            }
        }

        private void showAtoms() {
            List<Atom> atoms = SuperController.singleton.GetAtoms();
            foreach( Atom atom in atoms ) {
                if( (atom.type == _uiButton || atom.type == _uiSlider || atom.type == _uiToggle) ) {
                    var col = atom.GetStorableByID( "control" ) as FreeControllerV3;
                    if( col != null )
                        col.MoveYPositionRelativeNoForce( -10000 );
                }
            }
        }

        private List<NamedAtom> getAtoms() {
            List<Atom> atoms = SuperController.singleton.GetAtoms();
            List<NamedAtom> r = new List<NamedAtom>();
            int n = 1;
            HashSet<string> list = new HashSet<string>( StringComparer.OrdinalIgnoreCase );

            foreach( Atom atom in atoms ) {
                if( (atom.type == _uiButton || atom.type == _uiSlider || atom.type == _uiToggle)
                    && 
                    atom.on && atom.isActiveAndEnabled && atom.gameObject.activeInHierarchy ) {

                    if( _blackList.Contains( atom.uid ) ) continue;

                    var a = atom.GetStorableByID( "Text" ) as TextStorable;
                    var t = a.displayField.text.Trim( '\n', '\r', ' ', '\t' )
                        .Replace( '\n', ' ' )
                        .Replace( "  ", " " );

                    if( string.IsNullOrEmpty( t ) || string.Compare( t, "button", true ) == 0 ) {
                        t = atom.name.Trim( '\n', '\r', ' ', '\t' )
                        .Replace( '\n', ' ' )
                        .Replace( "  ", " " );
                    }
                    if( string.IsNullOrEmpty( t ) )
                        continue;

                    if( list.Contains( t ) ) {
                        t += n;
                        n++;
                    }
                    list.Add( t );

                    var trigger = atom.GetStorableByID( _storeTriggerId );
                    if( trigger != null && trigger.isActiveAndEnabled )
                        r.Add( new NamedAtom( t, atom ) );
                }
            }

            r.Sort( ( a, b ) => string.Compare( a.name, b.name ) );

            return r;
        }

        private Dictionary<string, string> getButtonAtoms() {
            List<Atom> atoms = SuperController.singleton.GetAtoms();
            Dictionary<string, string> d = new Dictionary<string, string>();

            foreach( Atom atom in atoms ) {
                if( (atom.type == _uiButton || atom.type == _uiToggle || atom.type == _uiSlider)
                    &&
                    atom.on && atom.isActiveAndEnabled && atom.gameObject.activeInHierarchy ) {

                    if( _blackList.Contains( atom.uid ) ) continue;

                    var a = atom.GetStorableByID( "Text" ) as TextStorable;
                    var t = a.displayField.text.Trim( '\n', '\r', ' ', '\t' )
                        .Replace( '\n', ' ' )
                        .Replace( "  ", " " );

                    if( string.IsNullOrEmpty( t ) || string.Compare( t, "button", true ) == 0 ) {
                        t = atom.name.Trim( '\n', '\r', ' ', '\t' )
                        .Replace( '\n', ' ' )
                        .Replace( "  ", " " );
                    }

                    t = t.ToLower();

                    if( d.ContainsKey( t ) )
                        continue;
                    if( string.IsNullOrEmpty( t ) )
                        continue;

                    var trigger = atom.GetStorableByID( _storeTriggerId );
                    if( trigger != null && trigger.isActiveAndEnabled ) {
                        if( atom.type == _uiSlider ) {
                            d[t + " up"] = atom.uid+"++";
                            d[t + " down"] = atom.uid + "--";
                            d[t + " maximum"] = atom.uid + "**";
                            d[t + " minimum"] = atom.uid + "__";
                        } else {
                            d[t] = atom.uid;
                        }
                    }
                }
            }

            return d;
        }
        #endregion

        #region Save/Load
        static string getString<K>( Dictionary<K, string> dict ) {
            StringBuilder sb = new StringBuilder();
            foreach( var item in dict )
                sb.AppendFormat( "{0}={1}\n", item.Key, item.Value );
            return sb.ToString();
        }

        void setHotKeysFromString( string s ) {
            _hotKeys.Clear();
            try {
                string[] items = s.Split( new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries );
                foreach( var item in items ) {
                    var values = item.Split( '=' );
                    var val = (KeyCode)Enum.Parse( typeof(KeyCode), values[0] );
                    _hotKeys[val] = values[1];
                }
            } catch {
                _hotKeys.Clear();
            }
        }

        void setSpeechHotKeysFromString( string s ) {
            _speechHotKeys.Clear();
            try {
                string[] items = s.Split( new string[] { "\n" }, StringSplitOptions.RemoveEmptyEntries );
                foreach( var item in items ) {
                    var values = item.Split( '=' );
                    _speechHotKeys[values[0]] = values[1];
                }
            } catch {
                _speechHotKeys.Clear();
            }
        }

        void sceneSave() {
            if( !_sessionMode ) {
                _sceneHotKeys.val = getString( _hotKeys );
                _sceneSpeechHotKeys.val = getString( _speechHotKeys );
                _sceneBlacklist.val = string.Join( "|", _blackList.ToArray() );
                _isDirty = false;
            }
        }

        void saveNow() {
            if( _sessionMode ) {
                if( string.IsNullOrEmpty( SuperController.singleton.LoadedSceneName ) )
                    return;

                var json = loadJson() ?? new SimpleJSON.JSONClass();

                var data = new SimpleJSON.JSONClass {
                    ["hotKeys"] = getString( _hotKeys ),
                    ["speechHotKeys"] = getString( _speechHotKeys ),
                    ["blackList"] = string.Join( "|", _blackList.ToArray() )
                };
                json[SuperController.singleton.LoadedSceneName] = data;

                try {
                    SuperController.singleton.SaveJSON( json, $"Saves\\autohotkeys.json" );
                    _isDirty = false;
                } catch( Exception e ) {
                    SuperController.LogError( e.ToString(), false );
                }
            }
        }

        void load() {
            _isDirty = false;
            if( _sessionMode ) {
                var result = loadJson();
                try {
                    if( result != null && !string.IsNullOrEmpty( SuperController.singleton.LoadedSceneName ) ) {
                        var scene = result[SuperController.singleton.LoadedSceneName];
                        if( scene != null ) {
                            if( scene["hotKeys"] != null )
                                setHotKeysFromString( scene["hotKeys"].Value );
                            if( scene["speechHotKeys"] != null )
                                setSpeechHotKeysFromString( scene["speechHotKeys"].Value );
                            _blackList.Clear();
                            if( scene["blackList"] != null ) {
                                var blacks = scene["blackList"].Value.Split( '|' );
                                foreach( var black in blacks )
                                    _blackList.Add( black );
                            }

                            updateKeywordRecognizer();
                            return;
                        }
                    }
                } catch {
                }
                _speechHotKeys.Clear();
                _hotKeys.Clear();
                _blackList.Clear();

                if( result == null )
                    saveNow();
            } else {
                setHotKeysFromString( _sceneHotKeys.val );
                setSpeechHotKeysFromString( _sceneSpeechHotKeys.val );
                _blackList.Clear();
                var blacks = _sceneBlacklist.val.Split( '|' );
                foreach( var black in blacks )
                    _blackList.Add( black );
            }

            updateKeywordRecognizer();
        }

        static JSONClass loadJson() {
            try {
                var settings = SuperController.singleton.LoadJSON( $"Saves\\autohotkeys.json" );
                if( settings == null ) {
                    return null;
                }
                return settings.AsObject;
            } catch( Exception e ) {
                // note: can't load "System.IO" without error so we will ignore a file
                // not found error by it's text (sadface)
                if( e.Message.Contains( "Could not find file" ) ) {
                    // log nothing
                } else {
                    SuperController.LogError( e.ToString(), false );
                }
                return null;
            }
        }
        #endregion

        #region String comparison functions
        delegate bool CompareDelegate( string pattern, string text );

        static bool compare1( string pattern, string text ) {
            if( string.IsNullOrEmpty( pattern ) )
                return false;
            return text.ToLower().Contains( pattern.ToLower() );
        }

        static bool compare2( string pattern, string text ) {
            if( string.IsNullOrEmpty( pattern ) )
                return false;

            int start = 0;
            for( int j = 0; j < pattern.Length; ++j ) {
                if( start >= text.Length )
                    return false;

                start = text.IndexOf( pattern[j].ToString(), start, StringComparison.OrdinalIgnoreCase );

                if( start == -1 ) {
                    return false;
                } else {
                    start++;
                }
            }

            return true;
        }
        #endregion

        #region Helper classes and DTOs
        class Keys {
            static readonly KeyCode[] blackList = new KeyCode[] {
                KeyCode.None,
                KeyCode.F1,
                KeyCode.F11,
                KeyCode.F12,
                KeyCode.A,
                KeyCode.S,
                KeyCode.D,
                KeyCode.W,
                KeyCode.U,

                KeyCode.LeftAlt,
                KeyCode.LeftControl,
                KeyCode.LeftShift,
            };

            readonly bool[] _down;
            readonly KeyCode[] _keys;

            public Keys() {
                _keys = System.Enum.GetValues( typeof( KeyCode ) )
                        .Cast<KeyCode>()
                        .Where( k => ((int)k < (int)KeyCode.Mouse0) )
                        .ToArray();
                _down = new bool[(int)KeyCode.Mouse0];
            }

            public void Update() {
                Letter = null;
                FirstKey = KeyCode.None;

                for( int j = 0; j < _keys.Length; ++j ) {
                    var x = _down[(int)_keys[j]] = Input.GetKeyDown( _keys[j] );

                    if( x ) {
                        if( FirstKey == KeyCode.None && !blackList.Contains( _keys[j] ) )
                            FirstKey = _keys[j];

                        if( (_keys[j] >= KeyCode.A && _keys[j] <= KeyCode.Z)
                        ||
                        (_keys[j] >= KeyCode.Alpha0 && _keys[j] <= KeyCode.Alpha9) ) {

                            Letter = char.ConvertFromUtf32( (int)_keys[j] );
                        } else if( _keys[j] == KeyCode.Period ) {
                            Letter = ">";
                        } else if( _keys[j] == KeyCode.Comma ) {
                            Letter = "<";
                        }
                    }
                }
            }

            public bool IsPressed( KeyCode key ) {
                return _down[(int)key];
            }

            public string Letter { get; private set; }

            //public KeyCode HotHotKey { get; private set; }

            public KeyCode FirstKey { get; private set; }
        }

        class NamedAtom {
            public readonly string name;
            public readonly Atom atom;
            public readonly string postfix;

            public NamedAtom( string name, Atom atom ) {
                this.name = name;
                this.atom = atom;

                switch( atom.type ) {
                    //case "UIButton":
                    //    this.postfix = "";
                    //    break;
                    case _uiSlider:
                        this.postfix = $"{(atom.GetStorableByID( _storeTriggerId ) as UISliderTrigger).slider.value:0.0}";
                        break;
                    case _uiToggle:
                        this.postfix = $"{((atom.GetStorableByID( _storeTriggerId ) as UIToggleTrigger).toggle.isOn ? "on": "off")}";
                        break;
                }
            }
        }
        #endregion
    }
}